/*
 * emdros_wrapper.cpp
 *
 * A wrapper around EmdrosEnv.
 *
 * Ulrik Petersen
 * Created: 4/9-2005
 * Last update: 3/24-2008
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2005-2008  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

#include "emdros_wrapper.h"
#include "harvest.h"
#include "output_style.h"
#include "schema.h"
#include <emdf_output.h>
#include <sstream>

EmdrosWrapper::EmdrosWrapper(EmdrosEnv *pEE, 
			     Configuration *pConf,
			     QTOutputBase *pOut) 
{
	m_pEE = pEE;
	m_pConf = pConf;
	m_pOut = pOut;
	m_pSchema = 0;
}

EmdrosWrapper::~EmdrosWrapper()
{ 
	// You don't own the other members!
	delete m_pSchema; 
}

bool EmdrosWrapper::configurePOut(std::string& message)
{
	if (m_pEE != 0
	    && m_pConf != 0) {
		std::ostringstream ostrStream;
		if (!retrieveSchema(&ostrStream)) {
			message = ostrStream.str();
			return false;
		} else {
			return m_pOut->getConfigurationValues(m_pConf, m_pSchema, message);
		}
	} else {
		message = "EmdrosWrapper not configured. This is an internal error.";
		return false;
	}
}


bool EmdrosWrapper::executeString(const std::string& query) 
{
	bool bNoCompilerError = true;
	bool bNoDBError = m_pEE->executeString(query, bNoCompilerError, false, true);
	return bNoCompilerError && bNoDBError;
}


void EmdrosWrapper::show_result() 
{
	if (m_pEE->isTable()) {
		show_table(m_pEE->takeOverTable());
	} else if (m_pEE->isSheaf()) {
		show_sheaf(m_pEE->takeOverSheaf());
	} else {
		(*m_pOut) << "Query did not yield any results.";
		m_pOut->endl();
	}

	// Make sure that the results are displayed,
	// even if the "quiet" option is on.
	m_pOut->flush();
}

void EmdrosWrapper::show_table(Table *pTable) 
{
	std::ostringstream ostr;
	EMdFOutput emdf_output(kCSISO_8859_1, &ostr, kOKConsole);
	pTable->printConsole(&emdf_output);
	delete pTable;
	(*m_pOut) << ostr.str();
	printIfNotQuiet("\nDone printing table.");
}

void EmdrosWrapper::show_sheaf(Sheaf *pSheaf) 
{
	if (pSheaf->isFail()) {
		(*m_pOut) << "No solutions found.";
		m_pOut->endl();
	} else {
		printIfNotQuiet("Now harvesting results...");
		Harvest h(this, pSheaf);
		printIfNotQuiet("... done harvesting!");
		if (!m_pOut->ShouldStop()) {
			m_pEE->clean();
			std::string output_style = *(m_pConf->getValues("output_style").begin());
			if (output_style == "output") {
				Output_style outstyle(m_pConf, m_pOut);
				printIfNotQuiet("Number of solutions: " + long2string(h.size()));
				unsigned int i = 0;
				SolutionListConstIterator ci = h.const_iterator();
				while (ci.hasNext() && !m_pOut->ShouldStop()) {
					const Solution *pSolution = ci.next();
					outstyle.show(pSolution, i+1);
					++i;
				}
			} else if (output_style == "kwic") {
				KWIC_Output_style outstyle(m_pConf, m_pOut);
				printIfNotQuiet("Number of solutions: " + long2string(h.size()));
				m_pOut->startTable("kwic");
				unsigned int i = 0;
				SolutionListConstIterator ci = h.const_iterator();
				while (ci.hasNext() && !m_pOut->ShouldStop()) {
					const Solution *pSolution = ci.next();
					outstyle.show(pSolution, i+1);
					++i;
				}
				m_pOut->endTable("kwic");
			} else if (output_style == "tree") {
				Tree_Output_style outstyle(m_pConf, m_pOut);
				printIfNotQuiet("Number of solutions: " + long2string(h.size()));
				unsigned int i = 0;
				SolutionListConstIterator ci = h.const_iterator();
				while (ci.hasNext() && !m_pOut->ShouldStop()) {
					const Solution *pSolution = ci.next();
					outstyle.show(pSolution, i+1);
					++i;
				}
			} else {
				(*m_pOut) << "Unknown output_style: '" + output_style + "' in configuration file.";
				m_pOut->newline();
				(*m_pOut) << "The only valid option at the moment is \"output\" (without the quotes).";
				m_pOut->endl();
			}
			printIfNotQuiet("\nDone printing sheaf.");
		}
	}
	delete pSheaf;
}


void EmdrosWrapper::printIfNotQuiet(const std::string& toPrint)
{
	if (!m_pConf->has("option", "quiet")) {
		if (m_pOut != 0) {
			m_pOut->startLatinParagraph();
			(*m_pOut) << toPrint;
			m_pOut->endl();
		}
	}
}


bool EmdrosWrapper::retrieveSchema(std::ostream *pOut)
{
	if (!m_pEE->connectionOk()) {
		(*pOut) << "Connection to backend not OK.";
		return false;
	} else {
		delete m_pSchema;
		m_pSchema = new EmdrosSchema(this->m_pEE);
		if (!m_pSchema->populate(pOut)) {
			delete m_pSchema;
			m_pSchema = 0;
			return false; 
		} else {
			return true;
		}
	}
}
